'''macros.py: Generate macro values from configuration values and provide
substitution functions.

The following macros are available:

  LCODE CCODE PCCODE PKGNAME PKGCODE LANG COUNTRY PCOUNTRY DISTRO RELEASE UPLOADER
  DATE TIMESTAMP CLASS CLASSNAME CLASSNAMESPACE CLASSDEPENDS
  BASERECOMMENDS
'''

import time
import re
import os
from decimal import Decimal


def _file_map(file, key, sep=None):
    '''Look up key in given file ("key value" lines).

    Throw KeyError if key was not found.
    '''
    val = None
    with open(file) as f:
        for l in f:
            try:
                (k, v) = l.split(sep, 1)
            except ValueError:
                continue
            # sort out comments
            if k.find('#') >= 0 or v.find('#') >= 0:
                continue
            if k == key:
                val = v.strip()
    if val is None:
        raise KeyError('Key %s not found in %s' % (key, file))
    return val


class LangpackMacros:
    def __init__(self, distro, locale, cls, release, version):
        '''Initialize values of macros.

        This uses information from maps/, config/, some hardcoded aggregate
        strings (such as package names), and some external input:

        - locale: Standard locale representation (e. g. pt_BR.UTF-8)
        - cls: empty, or valid name in maps/classnames
        - release: distrorelease name (in maps/releaseversions-$distro)
        - version: export date of language pack tarball, will be used in
          version numbers
        '''
        self.macros = {}
        # chop .* and @* suffixes to get encoding-agnostic locale
        try:
            self['LOCALE'] = (locale.split('.')[0]).split('@')[0]
        except KeyError:
            self['LOCALE'] = locale

        # class
        self['CLASS'] = cls
        if cls:
            self['CLASSNAME'] = _file_map('maps/classnames', cls)
            self['CLASSNAMESPACE'] = self['CLASSNAME'] + ' '

        # language and country
        try:
            (self['LCODE'], self['CCODE']) = self['LOCALE'].split('_')
            self['PCCODE'] = '(%s)' % self['CCODE']
            try:
                self['COUNTRY'] = _file_map('maps/countries', self['CCODE'], ':')
                self['PCOUNTRY'] = '(%s)' % self['COUNTRY']
            except KeyError:
                self['COUNTRY'] = ''
                self['PCOUNTRY'] = ''
        except ValueError:
            self['LCODE'] = self['LOCALE']

        if not self['LCODE']:
            raise Exception('Internal error: LCODE is empty')

        # other variables
        self['DATE'] = time.strftime('%a, %d %b %Y %H:%M:%S +0000')
        self['DISTRO'] = distro
        self['RELEASE'] = release
        self['RELEASEVERSION'] = _file_map('maps/releaseversions-' + distro, release.split('-')[0])
        self['TIMESTAMP'] = version

        # package name
        if Decimal(self['RELEASEVERSION']) >= Decimal('9.10'):
            try:
                self['PKGCODE'] = _file_map('maps/locale2pkgname', '%s-%s' % (self['LCODE'], self['CCODE'].lower()), ':')
                self['LANG'] = _file_map('maps/variants', '%s' % self['PKGCODE'], ':')
                if self['CLASS']:
                    self['PKGNAME'] = '%s-%s' % (self['CLASS'], self['PKGCODE'])
                else:
                    self['PKGNAME'] = self['PKGCODE']
            except KeyError:
                self['PKGCODE'] = self['LCODE']
                try:
                    self['LANG'] = _file_map('maps/variants', '%s' % self['PKGCODE'], ':')
                except KeyError:
                    self['LANG'] = _file_map('maps/languages', self['LCODE'], ':')
                if self['CLASS']:
                    self['PKGNAME'] = '%s-%s' % (self['CLASS'], self['LCODE'])
                else:
                    self['PKGNAME'] = self['LCODE']
        else:
            self['PKGCODE'] = self['LCODE']
            self['LANG'] = _file_map('maps/languages', self['LCODE'], ':')
            if self['CLASS']:
                self['PKGNAME'] = '%s-%s' % (self['CLASS'], self['LCODE'])
            else:
                self['PKGNAME'] = self['LCODE']

        # %CLASS% packages should depend on corresponding generic one
        if cls != '':
            self['CLASSDEPENDS'] = ', language-pack-%s' % self['PKGCODE']
        else:
            self['CLASSDEPENDS'] = ''

        # configuration files
        f = open('config/UPLOADER')
        self['UPLOADER'] = f.read().strip()
        f.close()

        # conflicts / replaces for zh langpack-split
        if self['PKGCODE'] == 'zh-hans' or self['PKGCODE'] == 'zh-hant':
            if cls != '':
                self['OBSOLETESZHBASE'] = ', language-pack-%s-zh-base (<< 1:9.10)' % self['CLASS']
                self['OBSOLETESZH'] = ', language-pack-%s-zh (<< 1:9.10)' % self['CLASS']
                self['CONFLICTSZH'] = '\nConflicts: language-pack-%s-zh (<< 1:9.10)' % self['CLASS']
            else:
                self['OBSOLETESZHBASE'] = ', language-pack-zh-base (<< 1:9.10)'
                self['OBSOLETESZH'] = ', language-pack-zh (<< 1:9.10)'
                self['CONFLICTSZH'] = '\nConflicts: language-pack-zh (<< 1:9.10)'
        else:
            self['OBSOLETESZHBASE'] = ''
            self['OBSOLETESZH'] = ''
            self['CONFLICTSZH'] = ''

    def __getitem__(self, item):
        # return empty string as default
        return self.macros.get(item, '')

    def __setitem__(self, item, value):
        self.macros[item] = value

    def __contains__(self, item):
        return self.macros.__contains__(item)

    def subst_string(self, s):
        '''Substitute all macros in given string.'''

        re_macro = re.compile('%([A-Z]+)%')
        while 1:
            m = re_macro.search(s)
            if m:
                s = s[:m.start(1) - 1] + self[m.group(1)] + s[m.end(1) + 1:]
            else:
                break

        return s

    def subst_file(self, file):
        '''Substitute all macros in given file.'''

        f = open(file)
        s = f.read()
        f.close()
        f = open(file, 'w')
        f.write(self.subst_string(s))
        f.close()

    def subst_tree(self, root):
        '''Substitute all macros in given directory tree.'''

        for path, dirs, files in os.walk(root):
            for f in files:
                self.subst_file(os.path.join(root, path, f))


if __name__ == '__main__':
    import tempfile
    import unittest

    class _T(unittest.TestCase):
        def test_lang(self):
            '''language, no country, no class'''

            l = LangpackMacros('ubuntu', 'de', '', 'precise', '3.14')
            self.assertEqual(l.subst_string('%LCODE% "%CCODE%" language-pack-%PKGNAME%'),
                             'de "" language-pack-de')
            self.assertEqual(l.subst_string('-%PCCODE%-%PKGCODE%-%LANG%-%RELEASE%'),
                             '--de-German-precise')
            self.assertEqual(l.subst_string('-%COUNTRY%-%PCOUNTRY%-'), '---')
            self.assertNotEqual(l.subst_string('%UPLOADER%'), '')
            self.assertTrue(len(l.subst_string('%DATE%')) > 10)
            self.assertEqual(l.subst_string('-%CLASS%-%CLASSNAME%-%CLASSNAMESPACE%-'), '----')
            self.assertEqual(l.subst_string('%CLASSDEPENDS%'), '')
            self.assertEqual(l.subst_string('%DISTRO%'), 'ubuntu')

        def test_lang_country(self):
            '''language, country, no class'''

            for locale in ('de_CH', 'de_CH.utf8'):
                l = LangpackMacros('ubuntu', locale, '', 'precise', '3.14')
                self.assertEqual(l.subst_string('%LCODE% "%CCODE%" language-pack-%PKGNAME%'),
                                 'de "CH" language-pack-de')
                self.assertEqual(l.subst_string('-%PCCODE%-%PKGCODE%-%LANG%-%RELEASE%'),
                                 '-(CH)-de-German-precise')
                self.assertEqual(l.subst_string('-%COUNTRY%-%PCOUNTRY%-'), '-Switzerland-(Switzerland)-')
                self.assertNotEqual(l.subst_string('%UPLOADER%'), '')
                self.assertTrue(len(l.subst_string('%DATE%')) > 10)
                self.assertEqual(l.subst_string('-%CLASS%-%CLASSNAME%-%CLASSNAMESPACE%-'), '----')
                self.assertEqual(l.subst_string('%CLASSDEPENDS%'), '')

        def test_lang_country_class(self):
            '''language, country, class'''

            for locale in ('de_CH', 'de_CH.utf8'):
                l = LangpackMacros('ubuntu', locale, 'gnome', 'trusty', '3.14')
                self.assertEqual(l.subst_string('%LCODE% "%CCODE%" language-pack-%PKGNAME%'),
                                 'de "CH" language-pack-gnome-de')
                self.assertEqual(l.subst_string('-%PCCODE%-%PKGCODE%-%LANG%-%RELEASE%'),
                                 '-(CH)-de-German-trusty')
                self.assertEqual(l.subst_string('-%COUNTRY%-%PCOUNTRY%-'), '-Switzerland-(Switzerland)-')
                self.assertNotEqual(l.subst_string('%UPLOADER%'), '')
                self.assertTrue(len(l.subst_string('%DATE%')) > 10)
                self.assertEqual(l.subst_string('-%CLASS%-%CLASSNAME%-%CLASSNAMESPACE%-'),
                                 '-gnome-GNOME-GNOME -')
                self.assertEqual(l.subst_string('%CLASSDEPENDS%'), ', language-pack-de')
                self.assertEqual(l.subst_string('%BASERECOMMENDS%'), '')

        def test_zh_split(self):
            '''Chinese split (PKGCODE)'''

            l = LangpackMacros('ubuntu', 'zh_CN', 'kde', 'trusty', '3.14')
            self.assertEqual(l.subst_string('%LCODE% "%CCODE%" language-pack-%PKGNAME%'),
                             'zh "CN" language-pack-kde-zh-hans')
            self.assertEqual(l.subst_string('-%PCCODE%-%PKGCODE%-%LANG%-%RELEASE%'),
                             '-(CN)-zh-hans-Simplified Chinese-trusty')
            self.assertEqual(l.subst_string('-%COUNTRY%-%PCOUNTRY%-'), '-China-(China)-')
            self.assertNotEqual(l.subst_string('%UPLOADER%'), '')
            self.assertTrue(len(l.subst_string('%DATE%')) > 10)
            self.assertEqual(l.subst_string('-%CLASS%-%CLASSNAME%-%CLASSNAMESPACE%-'),
                             '-kde-KDE-KDE -')
            self.assertEqual(l.subst_string('%CLASSDEPENDS%'), ', language-pack-zh-hans')

        def test_subst_file(self):
            '''subst_file()'''

            l = LangpackMacros('ubuntu', 'de', '', 'precise', '3.14')
            (fd, fname) = tempfile.mkstemp()
            try:
                os.close(fd)
                f = open(fname, 'w')
                f.write('%LCODE% "%CCODE%" language-pack-%PKGNAME%')
                f.close()
                l.subst_file(fname)
                with open(fname) as f:
                    self.assertEqual(f.read(), 'de "" language-pack-de')
            finally:
                os.unlink(fname)

        def test_unknown_broken(self):
            '''unknown macros and corner cases'''

            l = LangpackMacros('ubuntu', 'de', '', 'precise', '3.14')
            self.assertEqual(l.subst_string('%LCODE%% %%LANG% %%LCODE%% %LCODE LCODE%'),
                             'de% %German %de% %LCODE LCODE%')
            self.assertEqual(l.subst_string('%LCODE% %MOO% %LCODE%'),
                             'de  de')

    unittest.main()
